/*
** ocd - ojnk cd player
** Copyright (C) 1999 Odin <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <ocd.h>
#include <ctype.h>
#include <pwd.h>
#include <limits.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>

static u_int cddb_sum(u_int tlen);
static int cddb_login(const char *host, int port);
static int cddb_connect(const char *host, int port);
static int cddb_command(int d, u_char *buf, size_t len, const u_char *fmt, ...);
static int parse_repeat(const char *str, cddb_t *cddb);

extern WINDOW *cwin;
extern int max_x, max_y;

static const u_char *cddb_genre[] = {
	"blues",
	"classical",
	"country",
	"data",
	"folk",
	"jazz",
	"misc",
	"newage",
	"reggae",
	"rock",
	"soundtrack",
};

/* this assumes char [] and not char * for y */
#define FILLIN(x, y, z) {								\
	while (*x != '\0' && z < sizeof(y) - 1) {			\
		if (*x < 0x20 || (*x > 0x7e && *x < 0xa0)) {	\
			++x;										\
			continue;									\
		}												\
		if (*x == '\\') {								\
			++x;										\
			switch (*++x) {								\
				case 'n':								\
					y[z++] = '\n';						\
					break;								\
				case 't':								\
					y[z++] = '\t';						\
					break;								\
				case '\\':								\
					y[z++] = '\\';						\
					break;								\
				default:								\
					x++;								\
					break;								\
			}											\
		} else											\
			y[z++] = *x++;								\
	}													\
}

int cddb_lookup(struct cdrom *cd) {
	cd->cddb.discid = cddb_discid(cd->fd, &cd->disc);
	cd->cddb.tracks = cd->disc.tracks;
	cd->cddb.track = xcalloc(cd->disc.tracks, sizeof(struct cddb_track));

	if (cddb_setup(&cd->disc, &cd->cddb) == -1)
		return (-1);

	if (cd->cddb.playorder != NULL)
		return (1);

	return (0);
}

static u_int cddb_sum(u_int tlen) {
	int sum = 0;

	for (; tlen > 0 ; tlen /= 10) 
		sum += tlen % 10;

	return (sum);
}

u_long cddb_discid(int fd, const disc_t *d) {
	u_int t = 0, n = 0;
	struct ocd_time *offsets = xcalloc(d->tracks + 1, sizeof(struct ocd_time));

	if (ocd_get_offsets(fd, d->tracks, offsets) == -1)
		return (-1);

	for (; t < d->tracks ; t++)
		n += cddb_sum(offsets[t].min * 60 + offsets[t].sec);

	t = d->min * 60 + d->sec - offsets[0].min * 60 - offsets[0].sec;
	
	xfree(offsets);
	return (n % 0xff << 24 | t << 8 | d->tracks);
}

int cddb_setup(const disc_t *disc, cddb_t *cddb) {
	FILE *fp;
	u_char filename[PATH_MAX];

	snprintf(filename, sizeof(filename), "%s/%08lx", cddb->dir, cddb->discid);

	fp = fopen(filename, "r");
	if (fp == NULL) {
		fp = fopen(filename, "w+");
		if (fp == NULL)
			return (-1);
		cddb_unknown(fp, disc, cddb->discid);
		fseek(fp, 0, SEEK_SET);
	}

	return (cddb_read(fp, cddb));
}

static int cddb_connect(const char *host, int port) {
	int sock;

	mvwprintw(cwin, CDDB_STWIN_Y, CDDB_STWIN_X,
		"Connecting to %s:%u", host, port);
	wrefresh(cwin);

	sock = host_connect(host, port);
	if (sock == -1) {
		mvwprintw(cwin, CDDB_STWIN_Y, CDDB_STWIN_X,
			"Error: Could not connect to %s:%u", host, port);
		wrefresh(cwin);
		return (-1);
	}

	mvwprintw(cwin, CDDB_STWIN_Y, CDDB_STWIN_X,
		"Sucessfully connected to %s:%u", host, port);
	wrefresh(cwin);

	return (sock);
}

/*
** Log into a cddb server. (non http)
*/

static int cddb_login(const char *server, int port) {
	char buf[256];
	int code, sock = cddb_connect(server, port);

	if (sock == -1)
		return (-1);

	sock_readline(sock, NULL, 100);
	dprintf(sock, "proto %d\n", CDDB_PROTO);
	sock_readline(sock, NULL, 100);

	wmove(cwin, CDDB_STWIN_Y, CDDB_STWIN_X);
	wclrtoeol(cwin);

	/*
	** Is there a good reason to provide the server 
	** with meaningful information?
	*/
	code = cddb_command(sock, buf, sizeof(buf), "cddb hello O J N K\n");
	switch (code) {
		case CDDB_HS_OK:
			waddstr(cwin, "Handshake successful.");
			wrefresh(cwin);
			return (sock);
		case CDDB_HS_BD:
			waddstr(cwin, "Error: Handshake not successful.");
			break;
		case CDDB_HS_RE:
			waddstr(cwin, "Error: Already shook hands.");
			break;
		default:
			wprintw(cwin, "Error: Unknown error code. (%d)", code);
			break;
	}

	wrefresh(cwin);
	dprintf(sock, "quit\n");
	close(sock);

	return (-1);
}

int cddb_get(const disc_t *d, const cddb_t *cddb, FILE *fp, const char *server, int port) {
	u_char buf[256], genre[12] = { 0 }, *t;
	u_long i, discid = cddb->discid;
	int sock, code;

	sock = cddb_login(server, port);
	if (sock == -1)
		return (-1);

	dprintf(sock, "cddb query %08lx %u", cddb->discid, d->tracks);
	for (i = 0 ; i < d->tracks ; i++)
		dprintf(sock, " %u", d->track[i].start);

	wmove(cwin, CDDB_STWIN_Y, CDDB_STWIN_X);
	wclrtoeol(cwin);

	code = cddb_command(sock, buf, sizeof(buf), " %u\n", d->min * 60 + d->sec);
	switch (code) {
		case CDDB_EXACT:
			waddstr(cwin, "Found exact match.");
			wrefresh(cwin);
			break;
		case CDDB_MEXACT:
		case CDDB_INEXACT:
			if (code == CDDB_INEXACT)
				t = "Found inexact matches.";
			else
				t = "Found multiple exact matches.";

			waddstr(cwin, t);
			(void) wgetch(cwin);
			discid = cddb_inexact(sock, genre, sizeof(genre));
			wattrset(cwin, 0);
			cddb_window((cddb_t *) cddb);
			if (discid == 0)
				goto close_cddb;
			break;
		case CDDB_NOMATCH:
			waddstr(cwin, "Error: No match found.");
			goto close_cddb;
		case CDDB_DB_CORRUPT:
			waddstr(cwin, "Error: Database entry is corrupt.");
			goto close_cddb;
		case CDDB_HS_NO:
			waddstr(cwin, "Error: No handshake.");
			goto close_cddb;
		default:
			wprintw(cwin, "Error: Unknown error code. (%d)", code);
			goto close_cddb;
	}

	if (genre[0] == '\0') {
		t = strchr(&buf[4], ' ');
		if (t == NULL)
			goto close_cddb;
		*t = '\0';
		xstrncpy(genre, &buf[4], sizeof(genre));
	}

	wmove(cwin, CDDB_STWIN_Y, CDDB_STWIN_X);
	wclrtoeol(cwin);

	code = cddb_command(sock, buf, sizeof(buf), "cddb read %s %08lx\n",	genre,
						discid);
	switch (code) {
		case CDDB_FOUND:
			waddstr(cwin, "CDDB entry located.");
			wrefresh(cwin);
			break;
		case CDDB_NFOUND:
			waddstr(cwin, "Error: Specified CDDB entry not found.");
			goto close_cddb;
		case CDDB_SERVERR:
			waddstr(cwin, "Error: Server error.");
			goto close_cddb;
		case CDDB_LIMEX:
			waddstr(cwin, "Error: Access limit exceeded.");
			goto close_cddb;
		case CDDB_DB_CORRUPT:
			waddstr(cwin, "Error: Database entry is corrupt.");
			goto close_cddb;
		case CDDB_HS_NO:
			waddstr(cwin, "Error: No handshake.");
			goto close_cddb;
		case CDDB_BAD_SYNTAX:
			waddstr(cwin, "Error: Command syntax error.");
			goto close_cddb;
		default:
			wprintw(cwin, "Error: Unknown error code. (%d)", code);
			goto close_cddb;
	}

	for (i = 1 ; sock_readline(sock, buf, sizeof(buf)) ; i++) {
		if (buf[0] == '.')
			break;
		wclrtoeol(cwin);
		mvwprintw(cwin, CDDB_STWIN_Y, CDDB_STWIN_X,	"Read %lu lines.   ", i);
		wrefresh(cwin);
		fwrite(buf, sizeof(u_char), strlen(buf), fp);
	}

	mvwprintw(cwin, CDDB_STWIN_Y, CDDB_STWIN_X,
		"Download of CDDB database %08lx successful.", discid);

	wrefresh(cwin);
	dprintf(sock, "quit\n");
	close(sock);
	return (0);

close_cddb:
	wrefresh(cwin);
	dprintf(sock, "quit\n");
	close(sock);
	return (-1);
}

int cddb_submit(const cddb_t *cddb, const char *genre, const char *server, int port) {
	ssize_t len;
	struct stat st;
	char file[1024], buf[2048], *p, *q;
	int sock, fd;

	snprintf(file, sizeof(file), "%s/%08lx", cddb->dir, cddb->discid);

	wmove(cwin, CDDB_STWIN_Y, CDDB_STWIN_X);
	wclrtoeol(cwin);

	if (stat(file, &st) == -1) {
		waddstr(cwin, "Error: Could not stat cddb database file");
		goto bad;
	}

	if (st.st_size == 0) {
		waddstr(cwin, "Error: cddb database file is empty.");
		goto bad;
	}

	fd = open(file, O_RDONLY);

	if (fd == -1) {
		waddstr(cwin, "Error: Could not open cddb database file.");
		goto bad;
	}

	sock = cddb_connect(server, port);
	if (sock == -1) {
		close(fd);
		goto bad;
	}

	dprintf(sock, "POST /~cddb/submit.cgi HTTP/1.0\n");
	dprintf(sock, "Category: %s\n", genre);
	dprintf(sock, "Discid: %08lx\n", cddb->discid);
	dprintf(sock, "User-Email: user@example.com\n");
	dprintf(sock, "Submit-Mode: submit\n");
	dprintf(sock, "Content-Length: %lu\n\n", st.st_size);

	do {
		len = read(fd, buf, sizeof(buf));
		sock_write(sock, buf, len);
	} while (len == sizeof(buf));

	close(fd);
	wmove(cwin, CDDB_STWIN_Y, CDDB_STWIN_X);
	wclrtoeol(cwin);

	sock_readline(sock, buf, sizeof(buf));
	p = strchr(buf, ' ');
	if (p == NULL) {
		waddstr(cwin, "Error: Bad server output.");
		close(sock);
		goto bad;
	}
	q = strchr(++p, ' ');
	if (q == NULL) {
		waddstr(cwin, "Error: Bad server output.");
		close(sock);
		goto bad;
	}
	*q = '\0';
	len = atoi(p);

	switch (len) {
		case CDDB_SUBMIT_OK:
			waddstr(cwin, "OK, submission has been sent.");
			len = 0;
			break;
		case CDDB_SUBMIT_IE:
			waddstr(cwin, "Error: Internal error: failed to forward submission.");
			len = -1;
			break;
		case CDDB_SUBMIT_MI:
			waddstr(cwin, "Error: Missing required header information.");
			len = -1;
			break;
		default:
			wprintw(cwin, "Error: Unknown error code. (%d)", len);
			len = -1;
	}

	dprintf(sock, "quit\n");
	close(sock);
	wrefresh(cwin);

	return (len);
bad:
	wrefresh(cwin);
	return (-1);
}

void cddb_clear(cddb_t *cddb) {
	u_int i;

	memset(cddb->title, 0, sizeof(cddb->title));
	memset(cddb->ext, 0, sizeof(cddb->ext));

	for (i = 0 ; i < cddb->tracks ; i++) {
		memset(cddb->track[i].title, 0, sizeof(cddb->track[i].title));
		memset(cddb->track[i].ext, 0, sizeof(cddb->track[i].ext));
	}
}

int cddb_read(FILE *fp, cddb_t *cddb) {
	u_char *p, buf[128], pobuf[128] = { '\0' }, t;
	u_short ti_c = 0, ext_c = 0, po_c = 0;
	u_short *tr_c = xcalloc(cddb->tracks, sizeof(u_short));
	u_short *extt_c = xcalloc(cddb->tracks, sizeof(u_short));

	cddb_clear(cddb);
	
	while (ocd_getline(fp, buf, sizeof(buf) - 1) != EOF) {
		if (buf[0] == '#')
			continue;
		p = strchr(buf, '=');
		if (p == NULL)
			continue;
		*p++ = '\0';
		if (!strlen(p) || !strlen(buf))
			continue;
/*
**	As of my writing this, the valid cddb keywords are as follows:
**		DISCID, DTITLE, TTITLEN, EXTD, EXTTN, PLAYORDER
*/
		switch (buf[0]) {
			case 'D': /* We already have the DISCID */
				if (!strcmp(buf, "DTITLE")) {
					FILLIN(p, cddb->title, ti_c)
				}
				break;
			case 'T': /* Can only be TTITLEN */
				if (strncmp(buf, "TTITLE", 6))
					break;
				t = (u_char) atoi(&buf[6]);
				if (t >= cddb->tracks)
					break;
				FILLIN(p, cddb->track[t].title, tr_c[t]);
				break;
			case 'E':
				if (!strcmp(buf, "EXTD")) {
					FILLIN(p, cddb->ext, ext_c);
				} else if (!strncmp(buf, "EXTT", 4)) {
					t = (u_char) atoi(&buf[4]);
					if (t >= cddb->tracks)
						break;
					FILLIN(p, cddb->track[t].ext, extt_c[t]);
				}
				break;
			case 'P': /* Can only be PLAYORDER */
				if (!strcmp(buf, "PLAYORDER")) {
					FILLIN(p, pobuf, po_c);
				}
				break;
		}
	}

	for (ext_c = 0 , cddb->llen = 0 ; ext_c < cddb->tracks ; ext_c++) {
		if (tr_c[ext_c] > cddb->llen)
			cddb->llen = tr_c[ext_c];
	}

/*
** 13 is the number of chars before the first char of the song title... 
** (starting from the third horizontal cell (cell 2))
*/
	cddb->llen += 13;

	if (pobuf[0] != '\0') {
		if (parse_playlist(pobuf, cddb) == -1)
			destroy_playlist(cddb->playorder);
	}

	free(extt_c);
	free(tr_c);
	fclose(fp);
	return (0);
}

int cddb_valid_genre(const char *genre) {
	static size_t gsz = sizeof(cddb_genre) / sizeof(u_char *);

	return (astrcmp(genre, cddb_genre, gsz) == -1 ? 0 : 1);
}

int parse_playlist(u_char *p, cddb_t *cddb) {
	u_char *q, temp;

	if (cddb->playorder != NULL)
		destroy_playlist(cddb->playorder);

	for (; *p != '\0' ;) {
		if (!strncmp(p, "repeat(", 7)) {
			p += 7;
			q = strchr(p, ')');
			if (q == NULL) {
				/*
				** Malformed playlist repeat entry
				** fixme: print a nice warning
				*/
				destroy_playlist(cddb->playorder);
				cddb->playorder = NULL;
				return (-1);
			}
			*q = '\0';
			if (parse_repeat(p, cddb) == -1) {
				destroy_playlist(cddb->playorder);
				cddb->playorder = NULL;
				return (-1);
			}
			p = q + 2;
			continue;
		}
		q = p;
		p = strchr(q, ',');
		temp = (u_char) atoi(q);

		if (temp == 0) {
			/* Invalid data */
			destroy_playlist(cddb->playorder);
			cddb->playorder = NULL;
			return (-1);
		}

		if (p == NULL) {
			append_node(temp, &cddb->playorder);
			break;
		}

		*p++ = '\0';
		append_node(temp, &cddb->playorder);
	}

	while (cddb->playorder->prev != NULL)
		cddb->playorder = cddb->playorder->prev;

	return (0);
}

static int parse_repeat(const char *str, cddb_t *cddb) {
	char *p, buf[12];
	u_char i = 0, track = 0, times = 0;
	
	p = strchr(str, '{');
	if (p == NULL) {
		while (isspace(*str))
			str++;
		if (*str == '\0')
			return (-1);
		while (isdigit(*str)) 
			buf[i++] = *str++;
		buf[i] = '\0';
		if (i < 1 || *str == '\0')
			return (-1);
		track = (u_char) strtoul(buf, NULL, 10);
		if (track == 0)
			return (-1);
		while (isspace(*str))
			str++;
		if (*str != ',')
			return (-1);
		while (isspace(*++str))
			str++;
		for (i = 0 ; isdigit(*str) ;)
			buf[i++] = *str++;
		buf[i] = '\0';
		if (i < 1)
			return (-1);
		times = (u_char) strtoul(buf, NULL, 10);
		if (times == 0)
			return (-1);
		for (i = 0 ; i < times ; i++)
			append_node(track, &cddb->playorder);
	} else {
		u_char tbuf[128], c = 0, z = 0;
		char *q = strchr(p, '}');
		if (q == NULL)
			return (-1);
		*p++ = '\0';
		*q = '\0';

		for (i = 0 ; *p != '\0' ; ) {
			while (isspace(*p))
				p++;
			while (isdigit(*p))
				buf[i++] = *p++;
			/* might want to warn here */
			if (c >= sizeof(tbuf))
				break;
			if (i > 0) {
				buf[i] = '\0';
				tbuf[c] = (u_char) strtoul(buf, NULL, 10);
				if (tbuf[c++] == 0)
					return (-1);
				while (isspace(*p))
					p++;
				if (*p == '\0')
					break;
				if (*p != ',')
					return (-1);
				p++;
				i = 0;
				continue;
			} else {
				if (*p != '\0')
					break;
				return (-1);
			}
		}
		q++;
		while (isspace(*q))
			q++;
		if (*q++ != ',')
			return (-1);
		for (i = 0 ; isdigit(*q) ;)
			buf[i++] = *q++;
		if (i < 1)
			return (-1);
		buf[i] = '\0';
		times = (u_char) strtoul(buf, NULL, 10);	
		if (times == 0)
			return (-1);
		for (i = 0 ; i < times ; i++) {
			for (z = 0 ; z < c ; z++)
				append_node(tbuf[z], &cddb->playorder);
		}
	}

	return (0);
}

/* Feed the server a command, return the server return code */
static int cddb_command(int sock, u_char *buf, size_t len, const u_char *fmt, ...) {
	va_list ap;

	va_start(ap, fmt);
	vdprintf(sock, fmt, ap);
	va_end(ap);

	sock_readline(sock, buf, len);
	
	if (buf == NULL)
		return (0);

	while (isspace(*buf))
		++buf;
	*(buf + 3) = '\0';	

	return (atoi(buf));
}

void cddb_unknown(FILE *fp, const disc_t *d, u_long discid) {
	u_int i;

	fprintf(fp, "# xmcd CD Database Entry\n#\n# Track frame offsets:\n");
	
	for (i = 0 ; i < d->tracks ; i++)
		fprintf(fp, "# %u\n", d->track[i].start);

	fprintf(fp, "#\n# Disc length: %u seconds\n#\n# Revision: 0\n# Submitted via: %s\n#\nDISCID=%08lx\nDTITLE=Unknown / Unknown\n", d->min * 60 + d->sec, ME, discid);

	for (i = 0 ; i < d->tracks ; i++)
		fprintf(fp, "TTITLE%u=Track %u\n", i, i + 1);

	fprintf(fp, "EXTD=\n");

	for (i = 0 ; i < d->tracks ; i++)
		fprintf(fp, "EXTT%u=\n", i);

	fprintf(fp, "PLAYORDER=\n");
}
