/*
** ocd - ojnk cd player
** Copyright (C) 1999 Odin <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <ocd.h>
#include <fcntl.h>
#include <sys/ioctl.h>

/*
** Allocates the space for the track info.
*/
struct track *ocd_alloctracks(int fd) {
	struct cdrom_tochdr cdth;
	if (ioctl(fd, CDROMREADTOCHDR, &cdth) == -1)
		return (NULL);

	return (xcalloc(cdth.cdth_trk1 + 1, sizeof(struct track)));
}

/*
** Open the cdrom device, allocate the space for the track info.
*/
int ocd_setup(const char *device, disc_t *disc) {
	int fd;
	
	fd = open(device, O_RDONLY);
	if (fd == -1) {
		fprintf(stderr, "Error: %s: %m\n", device);
		return (-1);
	}

	disc->track = ocd_alloctracks(fd);
	if (disc->track == NULL) {
		close(fd);
		fd = -1;
	}

	return (fd);
}

int ocd_stop(int fd) {
	return (ioctl(fd, CDROMSTOP));
}

int ocd_pause(int fd) {
	return (ioctl(fd, CDROMPAUSE));
}

int ocd_resume(int fd) {
	return (ioctl(fd, CDROMRESUME));
}

int ocd_eject(int fd) {
	return (ioctl(fd, CDROMEJECT));
}

int ocd_close(int fd) {
	return (ioctl(fd, CDROMCLOSETRAY));
}

int ocd_getvol(int fd, struct cdrom_volctrl *volume) {
	return (ioctl(fd, CDROMVOLREAD, volume));
}

int ocd_setvol(int fd, const struct cdrom_volctrl *volume) {
	return (ioctl(fd, CDROMVOLCTRL, volume));
}

/*
** Updates disc information (disc time, track time, current track, status)
*/
int ocd_update(int fd, disc_t *disc) {
	u_int seconds;
	struct cdrom_subchnl cdsc;

	cdsc.cdsc_format = CDROM_MSF;

	if (ioctl(fd, CDROMSUBCHNL, &cdsc) == -1) {
		disc->present = 0;
		return (-1);
	}

	/* Hack to make eject work correctly */
	if (disc->opened == 1) {
		disc->present = 1;
		return (1);
	}

	disc->status = cdsc.cdsc_audiostatus;
	disc->pos.min = cdsc.cdsc_absaddr.msf.minute;
	disc->pos.sec = cdsc.cdsc_absaddr.msf.second;
	disc->frame = (cdsc.cdsc_absaddr.msf.minute * 60 + cdsc.cdsc_absaddr.msf.second) * 75 + cdsc.cdsc_absaddr.msf.frame;

	disc->c_track = 0;
	while (disc->c_track < disc->tracks && disc->frame >= disc->track[disc->c_track].start)
		disc->c_track++;

	if (disc->c_track == 0)
		return (0);

	seconds = (disc->frame - disc->track[disc->c_track - 1].start) / 75;
	disc->t_pos.min = seconds / 60;
	disc->t_pos.sec = seconds % 60;

	return (0);
}

/*
** Gather information about the disc that never changes (length, number
** of tracks, track lengths)
*/
int ocd_cdinfo(int fd, disc_t *disc) {
	struct cdrom_tochdr cdth;
	struct cdrom_tocentry cdte;
	u_char oldmin = 0, oldsec = 0;
	u_int i, j;

	if (ioctl(fd, CDROMREADTOCHDR, &cdth) == -1) {
		disc->present = 0;
		return (-1);
	}

	disc->tracks = cdth.cdth_trk1;
	cdte.cdte_format = CDROM_MSF;

	for (i = 0 ; i <= cdth.cdth_trk1 ; i++) {
		cdte.cdte_track = i == cdth.cdth_trk1 ? CDROM_LEADOUT : i + 1;
		
		if (ioctl(fd, CDROMREADTOCENTRY, &cdte) == -1) {
			disc->present = 0;
			return (-1);
		}

		disc->track[i].start = (cdte.cdte_addr.msf.minute * 60 + cdte.cdte_addr.msf.second) * 75 + cdte.cdte_addr.msf.frame;

		if (cdte.cdte_ctrl == CDROM_DATA_TRACK)
			disc->track[i].data = 1;

		if (i == 0) {
			oldmin = cdte.cdte_addr.msf.minute;
			oldsec = cdte.cdte_addr.msf.second;
			continue;
		}

		j = (cdte.cdte_addr.msf.minute * 60 + cdte.cdte_addr.msf.second) - (oldmin * 60 + oldsec);

		disc->track[i - 1].min = j / 60;
		disc->track[i - 1].sec = j % 60;

		oldmin = cdte.cdte_addr.msf.minute;
		oldsec = cdte.cdte_addr.msf.second;
	}

	disc->min = cdte.cdte_addr.msf.minute;
	disc->sec = cdte.cdte_addr.msf.second;

	disc->present = 1;
	return (0);
}

/*
** The cddb_discid function needs this information..
*/
int ocd_get_offsets(int fd, u_char tracks, struct ocd_time *offsets) {
	u_int i;
	struct cdrom_tocentry cdte;

	cdte.cdte_format = CDROM_MSF;
	for (i = 0 ; i <= tracks ; i++) {
		cdte.cdte_track = i == tracks ? CDROM_LEADOUT : i + 1;

		if (ioctl(fd, CDROMREADTOCENTRY, &cdte) == -1)
			return (-1);

		offsets[i].min = cdte.cdte_addr.msf.minute;
		offsets[i].sec = cdte.cdte_addr.msf.second;
	}

	return (0);
}

int ocd_play_track_pos(int fd, u_char start, u_char end, u_int startf) {
	u_int start_frame, end_frame;
	struct cdrom_msf cdmsf;
	disc_t disc;

	disc.track = ocd_alloctracks(fd);

	if (ocd_cdinfo(fd, &disc) == -1)
		return (-1);
	if (disc.track[start - 1].data == 1)
		return (-1);

	start_frame = disc.track[start - 1].start + startf * 75;
	
	if (end >= disc.tracks)
		end_frame = (disc.min * 60 + disc.sec) * 75;
	else
		end_frame = disc.track[end].start - 1;

	free(disc.track);

	cdmsf.cdmsf_min0 = start_frame / (60 * 75);
	cdmsf.cdmsf_sec0 = (start_frame % (60 * 75)) / 75;
	cdmsf.cdmsf_frame0 = start_frame % 75;
	cdmsf.cdmsf_min1 = end_frame / (60 * 75);
	cdmsf.cdmsf_sec1 = (end_frame % (60 * 75)) / 75;
	cdmsf.cdmsf_frame1 = end_frame % 75;

	return (ioctl(fd, CDROMPLAYMSF, &cdmsf));
}

int ocd_play_track(int fd, u_char track) {
	return (ocd_play_track_pos(fd, track, track, 0));
}

int ocd_play_tracks(int fd, u_char start, u_char end) {
	return (ocd_play_track_pos(fd, start, end, 0));
}

int ocd_play_pos(int fd, u_char track, u_int start) {
	struct cdrom_tochdr cdth;

	if (ioctl(fd, CDROMREADTOCHDR, &cdth) == -1)
		return (-1);

	return (ocd_play_track_pos(fd, track, cdth.cdth_trk1, start));
}

int ocd_play(int fd, u_char track) {
	struct cdrom_tochdr cdth;

	if (ioctl(fd, CDROMREADTOCHDR, &cdth) == -1)
		return (-1);

	return (ocd_play_tracks(fd, track, cdth.cdth_trk1));
}
